"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.deserializePayload = exports.serializePayload = exports.deserialize = exports.serialize = exports.registerPayloadType = exports.create = void 0;
const sdk_base_1 = require("@wormhole-foundation/sdk-base");
const layout_items_1 = require("./layout-items");
const utils_1 = require("./utils");
const uint8ArrayConversion = {
    to: (val) => val,
    from: (val) => val,
};
const guardianSignatureLayout = [
    { name: "guardianIndex", binary: "uint", size: 1 },
    { name: "signature", ...layout_items_1.signatureItem },
];
const headerLayout = [
    { name: "version", binary: "uint", size: 1, custom: 1, omit: true },
    { name: "guardianSet", ...layout_items_1.guardianSetItem },
    {
        name: "signatures",
        binary: "array",
        lengthSize: 1,
        layout: guardianSignatureLayout,
    },
];
const envelopeLayout = [
    { name: "timestamp", binary: "uint", size: 4 },
    { name: "nonce", binary: "uint", size: 4 },
    { name: "emitterChain", ...(0, layout_items_1.chainItem)() },
    { name: "emitterAddress", ...layout_items_1.universalAddressItem },
    { name: "sequence", ...layout_items_1.sequenceItem },
    { name: "consistencyLevel", binary: "uint", size: 1 },
];
const baseLayout = [...headerLayout, ...envelopeLayout];
// type govAA = "CoreBridgeUpgradeContract";
// type MyDescription = DescriptionOf<govAA>
// type MyType = LayoutToType<DescriptionOf<govAA>>;
// type MyVaa = { [K in keyof VAA<govAA>]: VAA<govAA>[K] };
// type Look = MyVaa["signatures"];
// type x = { [K in keyof VAA<"CoreBridgeUpgradeContract">]: VAA<"CoreBridgeUpgradeContract">[K] };
const payloadFactory = new Map();
function getPayloadDescription(payloadLiteral) {
    const description = payloadFactory.get(payloadLiteral);
    if (!description)
        throw new Error(`No layout registered for payload type ${payloadLiteral}`);
    return description;
}
const isCustomConversion = (val) => val.to !== undefined;
const descriptionToPayloadItem = (description) => (isCustomConversion(description)
    ? { name: "payload", binary: "bytes", custom: description }
    : {
        name: "payload",
        binary: "object",
        layout: description,
    });
const bodyLayout = (payloadLiteral) => [
    ...envelopeLayout,
    descriptionToPayloadItem(getPayloadDescription(payloadLiteral)),
];
const create = (payloadLiteral, vaaData) => {
    const body = bodyLayout(payloadLiteral);
    const bodyWithFixed = (0, sdk_base_1.addFixedValues)(body, 
    //not sure why the unknown cast here is required and why the type isn't inferred correctly
    vaaData);
    return {
        payloadLiteral,
        ...(0, sdk_base_1.addFixedValues)(headerLayout, vaaData),
        ...bodyWithFixed,
        hash: (0, utils_1.keccak256)((0, sdk_base_1.serializeLayout)(body, bodyWithFixed)),
    };
};
exports.create = create;
function registerPayloadType(payloadLiteral, payloadSerDe) {
    if (payloadFactory.has(payloadLiteral))
        throw new Error(`Payload type ${payloadLiteral} already registered`);
    payloadFactory.set(payloadLiteral, payloadSerDe);
}
exports.registerPayloadType = registerPayloadType;
const serialize = (vaa) => {
    const layout = [
        ...baseLayout,
        descriptionToPayloadItem(getPayloadDescription(vaa.payloadLiteral)),
    ];
    return (0, sdk_base_1.serializeLayout)(layout, vaa);
};
exports.serialize = serialize;
function deserialize(payloadLiteral, data) {
    if (typeof data === "string")
        data = (0, sdk_base_1.hexByteStringToUint8Array)(data);
    const [header, bodyOffset] = (0, sdk_base_1.deserializeLayout)(headerLayout, data, 0, false);
    //ensure that guardian signature indicies are unique and in ascending order - see:
    //https://github.com/wormhole-foundation/wormhole/blob/8e0cf4c31f39b5ba06b0f6cdb6e690d3adf3d6a3/ethereum/contracts/Messages.sol#L121
    for (let i = 1; i < header.signatures.length; ++i)
        if (header.signatures[i].guardianIndex <=
            header.signatures[i - 1].guardianIndex)
            throw new Error("Guardian signatures must be in ascending order of guardian set index");
    const body = (0, sdk_base_1.deserializeLayout)(bodyLayout(payloadLiteral), data, bodyOffset);
    const hash = (0, utils_1.keccak256)(data.slice(bodyOffset));
    return { payloadLiteral, ...header, ...body, hash };
}
exports.deserialize = deserialize;
const serializePayload = (payloadLiteral, payload) => {
    const description = getPayloadDescription(payloadLiteral);
    return isCustomConversion(description)
        ? description.from(payload)
        : (0, sdk_base_1.serializeLayout)(description, payload);
};
exports.serializePayload = serializePayload;
const deserializePayload = (payloadLiteral, data) => {
    const description = getPayloadDescription(payloadLiteral);
    data = typeof data === "string" ? (0, sdk_base_1.hexByteStringToUint8Array)(data) : data;
    return isCustomConversion(description)
        ? description.to(data)
        : (0, sdk_base_1.deserializeLayout)(description, data);
};
exports.deserializePayload = deserializePayload;
payloadFactory.set("Uint8Array", uint8ArrayConversion);
//# sourceMappingURL=vaa.js.map