import { hexByteStringToUint8Array, serializeLayout, deserializeLayout, addFixedValues, } from "@wormhole-foundation/sdk-base";
import { chainItem, universalAddressItem, signatureItem, sequenceItem, guardianSetItem, } from "./layout-items";
import { keccak256 } from "./utils";
const uint8ArrayConversion = {
    to: (val) => val,
    from: (val) => val,
};
const guardianSignatureLayout = [
    { name: "guardianIndex", binary: "uint", size: 1 },
    { name: "signature", ...signatureItem },
];
const headerLayout = [
    { name: "version", binary: "uint", size: 1, custom: 1, omit: true },
    { name: "guardianSet", ...guardianSetItem },
    {
        name: "signatures",
        binary: "array",
        lengthSize: 1,
        layout: guardianSignatureLayout,
    },
];
const envelopeLayout = [
    { name: "timestamp", binary: "uint", size: 4 },
    { name: "nonce", binary: "uint", size: 4 },
    { name: "emitterChain", ...chainItem() },
    { name: "emitterAddress", ...universalAddressItem },
    { name: "sequence", ...sequenceItem },
    { name: "consistencyLevel", binary: "uint", size: 1 },
];
const baseLayout = [...headerLayout, ...envelopeLayout];
// type govAA = "CoreBridgeUpgradeContract";
// type MyDescription = DescriptionOf<govAA>
// type MyType = LayoutToType<DescriptionOf<govAA>>;
// type MyVaa = { [K in keyof VAA<govAA>]: VAA<govAA>[K] };
// type Look = MyVaa["signatures"];
// type x = { [K in keyof VAA<"CoreBridgeUpgradeContract">]: VAA<"CoreBridgeUpgradeContract">[K] };
const payloadFactory = new Map();
function getPayloadDescription(payloadLiteral) {
    const description = payloadFactory.get(payloadLiteral);
    if (!description)
        throw new Error(`No layout registered for payload type ${payloadLiteral}`);
    return description;
}
const isCustomConversion = (val) => val.to !== undefined;
const descriptionToPayloadItem = (description) => (isCustomConversion(description)
    ? { name: "payload", binary: "bytes", custom: description }
    : {
        name: "payload",
        binary: "object",
        layout: description,
    });
const bodyLayout = (payloadLiteral) => [
    ...envelopeLayout,
    descriptionToPayloadItem(getPayloadDescription(payloadLiteral)),
];
export const create = (payloadLiteral, vaaData) => {
    const body = bodyLayout(payloadLiteral);
    const bodyWithFixed = addFixedValues(body, 
    //not sure why the unknown cast here is required and why the type isn't inferred correctly
    vaaData);
    return {
        payloadLiteral,
        ...addFixedValues(headerLayout, vaaData),
        ...bodyWithFixed,
        hash: keccak256(serializeLayout(body, bodyWithFixed)),
    };
};
export function registerPayloadType(payloadLiteral, payloadSerDe) {
    if (payloadFactory.has(payloadLiteral))
        throw new Error(`Payload type ${payloadLiteral} already registered`);
    payloadFactory.set(payloadLiteral, payloadSerDe);
}
export const serialize = (vaa) => {
    const layout = [
        ...baseLayout,
        descriptionToPayloadItem(getPayloadDescription(vaa.payloadLiteral)),
    ];
    return serializeLayout(layout, vaa);
};
export function deserialize(payloadLiteral, data) {
    if (typeof data === "string")
        data = hexByteStringToUint8Array(data);
    const [header, bodyOffset] = deserializeLayout(headerLayout, data, 0, false);
    //ensure that guardian signature indicies are unique and in ascending order - see:
    //https://github.com/wormhole-foundation/wormhole/blob/8e0cf4c31f39b5ba06b0f6cdb6e690d3adf3d6a3/ethereum/contracts/Messages.sol#L121
    for (let i = 1; i < header.signatures.length; ++i)
        if (header.signatures[i].guardianIndex <=
            header.signatures[i - 1].guardianIndex)
            throw new Error("Guardian signatures must be in ascending order of guardian set index");
    const body = deserializeLayout(bodyLayout(payloadLiteral), data, bodyOffset);
    const hash = keccak256(data.slice(bodyOffset));
    return { payloadLiteral, ...header, ...body, hash };
}
export const serializePayload = (payloadLiteral, payload) => {
    const description = getPayloadDescription(payloadLiteral);
    return isCustomConversion(description)
        ? description.from(payload)
        : serializeLayout(description, payload);
};
export const deserializePayload = (payloadLiteral, data) => {
    const description = getPayloadDescription(payloadLiteral);
    data = typeof data === "string" ? hexByteStringToUint8Array(data) : data;
    return isCustomConversion(description)
        ? description.to(data)
        : deserializeLayout(description, data);
};
payloadFactory.set("Uint8Array", uint8ArrayConversion);
//# sourceMappingURL=vaa.js.map